/*****************************************************************************
 *  File        : UART.C                    Version : 1.0                    *
 *                                                                           *
 *  Descriptions: Functions / Driver for UART port                           *
 *                                                                           *
 *  Author      : Andrew Hui                Date : 20 March, 1998            *
 *****************************************************************************/

#include "include\pr31700s.h"
#include "include\pr31700c.h"
#include "include\uart.h"

void UartPrints(BYTE *c)
{
	while (*c)
	{
		UartPutc(UART_A,*c);
		c++;
    }
}

/*---------------------------------------------------------------------------
  Prototype :   int  UartInit(int port, int baud, char config[] );
  Purpose   :   initialize the UART port for serial communication
  Scope     :   ALL
  input     :   int  port       UART_A: port A,     UART_B: port B
                int  baud       e.g.  38400
                char config[]   No_of_bits, Parity, stop_bit : "8N1"
  return    :   N/A
---------------------------------------------------------------------------*/
void  UartInit(int port, int baud, char *config )
{
    /* config SIB clock, for all serial comm. in side CPU */
    CPU->REG_CLK_CTRL |= SIBMCLKDIR | ENSIBMCLK | SIBMCLKDIV(2);
    CPU->REG_CLK_CTRL &= ~CSERSEL;    /* set internal clock souce */

    /* config UART clock */
    if (port==UART_A) CPU->REG_CLK_CTRL |= CSERDIV(3) | ENCSERCLK | ENUARTACLK;
    if (port==UART_B) CPU->REG_CLK_CTRL |= CSERDIV(3) | ENCSERCLK | ENUARTBCLK;

    /* config UART port */
    if (port ==UART_A)  /* UART A */
    {
        CPU ->REG_UARTA_CTRL_1     &= ~DISTXD;

        /* no. of bits */
        if ((char)config[0]=='7')   CPU ->REG_UARTA_CTRL_1 |=  BIT_7 ;
        if ((char)config[0]=='8')   CPU ->REG_UARTA_CTRL_1 &= ~BIT_7 ;

        /* Parity bits */
        if ((char)config[1]=='E') CPU->REG_UARTA_CTRL_1 |=EVENPARITY | ENPARITY;
        if ((char)config[1]=='O')
        {
            CPU ->REG_UARTA_CTRL_1 |= ENPARITY;
            CPU ->REG_UARTA_CTRL_1 &= ~EVENPARITY;
        }
        if ((char)config[1]=='N')   CPU ->REG_UARTA_CTRL_1 &= ~ENPARITY;

        /* no. of stop bit */
        if((char)config[2]=='2')    CPU ->REG_UARTA_CTRL_1 |=  TWOSTOP;
        if((char)config[2]=='1')    CPU ->REG_UARTA_CTRL_1 &= ~TWOSTOP;

        CPU ->REG_UARTA_CTRL_2      = BAUD(baud);   /* baud rate */
        CPU ->REG_UARTA_DMA_CTRL_1  = 0;            /* DMA buf start Address */
        CPU ->REG_UARTA_DMA_CTRL_2  = 0;            /* DMA buf length */
        CPU ->REG_UARTA_CTRL_1     |= ENUART;       /* enable Uart A / B */

        while(~CPU ->REG_UARTA_CTRL_1 & UARTON) ;   /*wait until Uart Astable*/
    }
    else  /* UART B */
    {
        CPU ->REG_UARTB_CTRL_1     &= ~DISTXD;

        /* no. of bits */
        if ((char)config[0]=='7')   CPU ->REG_UARTB_CTRL_1 |=  BIT_7 ;
        if ((char)config[0]=='8')   CPU ->REG_UARTB_CTRL_1 &= ~BIT_7 ;

        /* Parity bits */
        if ((char)config[1]=='E')
            CPU ->REG_UARTB_CTRL_1 |= EVENPARITY | ENPARITY;
        if ((char)config[1]=='O')
        {
            CPU ->REG_UARTB_CTRL_1 |= ENPARITY;
            CPU ->REG_UARTB_CTRL_1 &= ~EVENPARITY;
        }
        if ((char)config[1]=='N')   CPU ->REG_UARTB_CTRL_1 &= ~ENPARITY;

        /* no. of stop bit */
        if((char)config[2]=='2')    CPU ->REG_UARTB_CTRL_1 |=  TWOSTOP;
        if((char)config[2]=='1')    CPU ->REG_UARTB_CTRL_1 &= ~TWOSTOP;

        CPU ->REG_UARTB_CTRL_2      = BAUD(baud);   /* baud rate */
        CPU ->REG_UARTB_DMA_CTRL_1  = 0;            /* DMA buf start Address */
        CPU ->REG_UARTB_DMA_CTRL_2  = 0;            /* DMA buf length */
        CPU ->REG_UARTB_CTRL_1     |= ENUART;       /* enable Uart B */

        while(~CPU ->REG_UARTB_CTRL_1 & UARTON) ;   /* wait until Uart stable */
    }
}


/*---------------------------------------------------------------------------
  Prototype :   int UartStatus(int port );
  Purpose   :   Check UART port status
  Scope     :   ALL
  input     :   int  port        UART_A: port A,    UART_B: port B
  return    :   UART Port Status
---------------------------------------------------------------------------*/
int UartStatus(int port)
{   if (port==UART_A)      return (CPU ->REG_UARTA_CTRL_1);
    if (port==UART_B)      return (CPU ->REG_UARTB_CTRL_1);
}


/*---------------------------------------------------------------------------
  Prototype :   void UartPrintf(int port, char *buffer );
  Purpose   :   print a string to serial port
  Scope     :   ALL
  input     :   int  port        UART_A: port A,    UART_B: port B
            :   char *buffer     output buffer
  return    :   N/A
---------------------------------------------------------------------------*/
void UartPrintf(int port, char *buffer)
{   int i=0;
    while (buffer[i])
       UartPutc (port,  (char) buffer[i++] ) ;
}


/*---------------------------------------------------------------------------
  Prototype :   char  UartGetc(int port );
  Purpose   :   read a received byte from serial port
  Scope     :   ALL
  input     :   int  port   : UART_A: port A,    UART_B: port B
  return    :   result
---------------------------------------------------------------------------*/
char UartGetc(int port)
{   char data;
    while (UartRxData(port,  &data)!=UART_SUCCESS);
    return (char) data;
}


/*---------------------------------------------------------------------------
  Prototype :   void  UartPutc(int port, int data );
  Purpose   :   transmit a byte out the serial port
  Scope     :   ALL
  input     :   int  port       0 / UART_A: port A,   1 / UART_B: port B
  return    :   N/A
---------------------------------------------------------------------------*/
void UartPutc(int port, char data)
{
//        unsigned int j=0;
        while (UartTxData(port, data)!=UART_SUCCESS)
        {
//                j++;
//                if (j > 60000)    break;
        }
}


/*---------------------------------------------------------------------------
  Prototype :   int UartRxData(int port, char *data)
  Purpose   :   Read data from UART port
  Scope     :   ALL
  input     :   int  port  :  UART_A: port A,     UART_B: port B
  return    :   UART_SUCCESS
                UART_BUF_EMPTY
---------------------------------------------------------------------------*/
int  UartRxData(int port, char *data)
{
    if (port==UART_A)
    {
        if (~CPU ->REG_UARTA_CTRL_1 & RXHOLDFULL)   /* receive buffer empty */
            return (UART_BUF_EMPTY);
        *data = (char)(CPU ->REG_UARTA_HLDG );
    }
    if (port==UART_B)
    {
        if (~CPU ->REG_UARTB_CTRL_1 & RXHOLDFULL)   /* receive buffer empty */
            return (UART_BUF_EMPTY);
        *data = (char)(CPU ->REG_UARTB_HLDG );
    }
//    Delay(0xFFF);
    return (UART_SUCCESS);
}


/*---------------------------------------------------------------------------
  Prototype :   int UartTxData(int port, char data)
  Purpose   :   Write data to UART port
  Scope     :   ALL
  input     :   int  port     UART_A: port A,   UART_B: port B
  return    :    UART_SUCCESS
                 UART_BUF_FULL
---------------------------------------------------------------------------*/
int  UartTxData(int port, char data)
{

        unsigned int i=0;
    if (port==UART_A)
    {
        if (~CPU ->REG_UARTA_CTRL_1 & EMPTY)        /* transmit buffer full */
            return (UART_BUF_FULL);
        CPU ->REG_UARTA_HLDG = data;
    }
    else /* port == UART_B */
    {
        if (~CPU ->REG_UARTB_CTRL_1 & EMPTY)        /* transmit buffer full */
            return (UART_BUF_FULL);
        CPU ->REG_UARTB_HLDG = data;

    }

    while (CPU ->REG_UARTA_CTRL_1 & EMPTY)
    {
        i++;
        if (i > 60000)    break;
        }
    return (UART_SUCCESS);

}


